// Bindless G-Buffer.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

// True if the application is running
bool running = true;

// The window
SDL_Window* pWindow = NULL;

// The width and height of the window
Uint32 width = 256;
Uint32 height = 256;

// The GL context
SDL_GLContext context  = 0;

// The bindless g-buffer renderer
Renderer renderer;

// Init
static bool Init()
{
    // Initialize SDL
    if (SDL_Init(SDL_INIT_EVERYTHING) < 0) {
        return false;
    }

    // Create an SDL window and renderer
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
    SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
    pWindow = SDL_CreateWindow("Bindless G-Buffer", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width, height, SDL_WINDOW_OPENGL);
    if (!pWindow) {
        return false;
    }

    // Create a OpenGL context and make it current
    context = SDL_GL_CreateContext(pWindow);
    if (context == 0) {
        return false;
    }

    // When doing performance analysis we want to go as fast as possible
    SDL_GL_SetSwapInterval(0);

    // Initialize GLEW
    if (glewInit() != GLEW_OK) {
        return false;
    }

    // Make sure bindless is supported
    if (!BindlessBuffer::IsSupported() ||
        !BindlessTexture::IsSupported() ||
        !BindlessMesh::IsSupported())
    {
        printf("Missing extension.  This application requires GL_NV_vertex_buffer_unified_memory, GL_NV_gpu_shader5, and GL_NV_bindless_texture");
        return false;
    }

    // Initialize the renderer
    if (!renderer.Init(width, height, 7, 7, 7)) {
        return false;
    }
 
    return true;
}

// OnEvent
static void OnEvent(const SDL_Event& event)
{
    switch (event.type) {
    case SDL_QUIT:
        running = false;
        break;
    case SDL_KEYUP:
        switch (event.key.keysym.sym) {
        case SDLK_ESCAPE:
            running = false;
            break;
        case SDLK_1:
            SDL_SetWindowTitle(pWindow, "Bindless G-Buffer - All Passes");
            renderer.SetRenderMode(Renderer::RM_DEFAULT);
            break;
        case SDLK_2:
            SDL_SetWindowTitle(pWindow, "Bindless G-Buffer - Generate Only");
            renderer.SetRenderMode(Renderer::RM_GEN_ONLY);
            break;
        case SDLK_3:
            SDL_SetWindowTitle(pWindow, "Bindless G-Buffer - Sort Only");
            renderer.SetRenderMode(Renderer::RM_SORT_ONLY);
            break;
        case SDLK_4:
            SDL_SetWindowTitle(pWindow, "Bindless G-Buffer - Shader Only");
            renderer.SetRenderMode(Renderer::RM_SHADE_ONLY);
            break;
        case SDLK_5:
            SDL_SetWindowTitle(pWindow, "Bindless G-Buffer - Generate & Sort Only");
            renderer.SetRenderMode(Renderer::RM_GEN_AND_SORT);
            break;
        case SDLK_6:
            SDL_SetWindowTitle(pWindow, "Bindless G-Buffer - Sort & Shade Only");
            renderer.SetRenderMode(Renderer::RM_SORT_AND_SHADE);
            break;
        default:
            // Ignored
            break;
        }
        break;
    default:
        // Ignore
        break;
    }
}

// OnLoopPreRender
static void OnLoopPreRender()
{
    renderer.PreRender();
}

// OnRender
static void OnRender()
{
    renderer.Render();

    SDL_GL_SwapWindow(pWindow);
}

// OnLoopPostRender
static void OnLoopPostRender()
{
    renderer.PostRender();
}

// OnCleanup
static void OnCleanup()
{
    SDL_GL_DeleteContext(context);
    SDL_DestroyWindow(pWindow);
    SDL_Quit();
}

#define INVALID_COMMAND_LINE "Invalid command line.  Use:  Bindless G-Buffer [-w width] [-h height]"

// ParseCommandLine
static int ParseCommandLine(int argc, _TCHAR* argv[])
{
    for (int arg = 1; arg < argc; arg++) {
        if (wcscmp(argv[arg], L"-w") == 0) {
            arg++;
            if (arg < argc) {
                swscanf_s(argv[arg], L"%u", &width);
            } else {
                printf(INVALID_COMMAND_LINE);
                return -1;
            }
            continue;
        }

        if (wcscmp(argv[arg], L"-h") == 0) {
            arg++;
            if (arg < argc) {
                swscanf_s(argv[arg], L"%u", &height);
            } else {
                printf(INVALID_COMMAND_LINE);
                return -1;
            }
            continue;
        }

        printf(INVALID_COMMAND_LINE);
        return -1;
    }

    return 0;
}

// Main
int APIENTRY _tWinMain(_In_ HINSTANCE hInstance,
                     _In_opt_ HINSTANCE hPrevInstance,
                     _In_ LPTSTR    lpCmdLine,
                     _In_ int       nCmdShow)
{
    if (ParseCommandLine(__argc, __targv) != 0) {
        return -1;
    }

    // Initialize
    if (!Init()) {
        return -1;
    }

    // Application loop
    while (running) {
        // Event loop
        SDL_Event Event = { 0 };
        while (SDL_PollEvent(&Event)) {
            OnEvent(Event);
        }

        OnLoopPreRender();
        OnRender();
        OnLoopPostRender();
    }

    // Cleanup
    OnCleanup();

    return 0;
}
