#version 420 core
#extension GL_NV_gpu_shader5 : require
#extension GL_NV_shader_buffer_load : require
#extension GL_ARB_explicit_uniform_location : enable
#extension GL_NV_bindless_texture : require

#define FRAG_COLOR	0

#define UNIFORM_GBUFFER_MEMORY_POOL 3
#define UNIFORM_GBUFFER_ALLOCATOR 4
#define UNIFORM_GBUFFER 5
#define UNIFORM_GBUFFER_MAX_NODES 6
#define UNIFORM_MATERIALS 7
#define UNIFORM_RT_DIMS 8

// Input from the vertex shader
in block
{
    vec3 eyePosition; // eye space position
    vec3 eyeNormal;  // eye space normal
    vec2 uv; // uv (the same for all textures)
    flat uint material; // material index
} In;

// Material structure
struct Material {
    vec4 diffuseColor; // base diffuse color
    vec4 specularColor; // base specular color, shininess is packed in w
    uint64_t diffuseTexture; // texture to modulate diffuse color
    uint64_t specularTexture; // texture to modulate specular color (not used)
    uint64_t normalTexture; // texture to generate a normal (not used)
};

// GBufferNode structure that is written by this shader and consumed by later stages
struct GBufferNode {
    vec3 eyePosition; // eye space position
    float alpha; // alpha value (put it here for better packing)
    vec3 eyeNormal; // eye space normal
    uint material; // material index (put it here for better packing)
    vec2 uv; // uv for texture accesses
    uint64_t nextAddress; // the next address in the list
};

// The base pointer to a block of buffer object memory that we can sub-allocate from
layout (location = UNIFORM_GBUFFER_MEMORY_POOL) uniform GBufferNode* gbufferMemoryPool;

// The offset of the next entry in the gbuffer 
layout (location = UNIFORM_GBUFFER_ALLOCATOR) uniform uint64_t* pGBufferAllocator;

// The "image" of head pointers... entry (width * y + x) is the head pointer for the list for fragment (x,y)
layout (location = UNIFORM_GBUFFER) uniform GBufferNode** gbuffer;

// The size of the gbuffer
layout (location = UNIFORM_GBUFFER_MAX_NODES) uniform uint64_t gbufferMaxNodes;

// The array of materials
layout (location = UNIFORM_MATERIALS) uniform Material* pMaterials;

// The width and height of the render target
layout (location = UNIFORM_RT_DIMS) uniform uvec2 rtDims;

// The output color
layout(location = FRAG_COLOR, index = 0) out vec4 Color;

// Allocates a gbuffer node full of garbage
uint64_t AllocGBufferNodeAddress()
{
    uint64_t index = atomicAdd(pGBufferAllocator, 1);

    // Discard out of range accesses
    if (index >= gbufferMaxNodes) {
        return 0;
    }

    GBufferNode* pNode = gbufferMemoryPool + index;
    return uint64_t(pNode);
}

// Populates a gbuffer node from fragment shader input
void PopulateGBufferNode(inout GBufferNode* pNode)
{
    // Process the material for this node so that we have an alpha value for the fragment
    Material* pMaterial = pMaterials + In.material;
    float alpha = pMaterial->diffuseColor.a;
    if (pMaterial->diffuseTexture != 0) {
        sampler2D tex = sampler2D(pMaterial->diffuseTexture);
        alpha *= texture(tex, In.uv).a;
    }

    // Write out properties of this node
    pNode->eyePosition = In.eyePosition;
    pNode->alpha = alpha;
    pNode->eyeNormal = In.eyeNormal;
    pNode->material = In.material;
    pNode->uv = In.uv;
    pNode->nextAddress = 0;
}

// Inserts a gbuffer node for the current fragment
void InsertGBufferNode(inout GBufferNode* pNode)
{
    // Setup a pointer to the head of the list
    uint64_t* pHeadAddress = (uint64_t*)gbuffer;
    pHeadAddress += (rtDims.x * uint(gl_FragCoord.y)) + uint(gl_FragCoord.x);

    // Insertion is done using an atomic exchange.  At this point in time
    // pHeadAddress points to something.  It could be NULL (initialization 
    // value), or it could be a node.  The insertion works by pointing pHead
    // at this node (atomically).  The exchange returns the old value, which 
    // can then be glued up to pNode->nextAddress.
    pNode->nextAddress = atomicExchange(pHeadAddress, uint64_t(pNode));
}

void main()
{
    // Try to allocate a new node
    uint64_t nodeAddress = AllocGBufferNodeAddress();
    if (nodeAddress == 0) {
        // Buffer overflow!  Discard once this works... draw red for now.
        discard;
        return;
    }

    // Populate the new node
    GBufferNode* pNode = (GBufferNode*)nodeAddress;
    PopulateGBufferNode(pNode);

    // Insert the new node into the list for this fragment
    InsertGBufferNode(pNode);

    // For debugging...
    // Write out something so we can see what is going on... uv is a good choice
    Color.rg = pNode->uv;
}