// Renderer.cpp

#include "stdafx.h"

#define UNIFORM_MVP_XFORM 0
#define UNIFORM_MV_XFORM 1
#define UNIFORM_NORMAL_XFORM 2
#define UNIFORM_GBUFFER_MEMORY_POOL 3
#define UNIFORM_GBUFFER_ALLOCATOR 4
#define UNIFORM_GBUFFER 5
#define UNIFORM_GBUFFER_MAX_NODES 6
#define UNIFORM_MATERIALS 7
#define UNIFORM_RT_DIMS 8
#define UNIFORM_CLEAR_COLOR 9
#define UNIFORM_AMBIENT_LIGHT_COLOR 10
#define UNIFORM_LIGHT_COUNT 11
#define UNIFORM_LIGHTS 12
#define UNIFORM_MATERIAL_IDS 13
#define UNIFORM_GBUFFER_ALLOCATOR_DIMS 14
#define UNIFORM_GBUFFER_ALLOCATOR_MAX_NODES_PER_TILE 15

// Tile width and height in pixels
#define ALLOCATOR_TILE_WIDTH 2
#define ALLOCATOR_TILE_HEIGHT 2

// Number of nodes per pixel (based on average depth complexity)
#define NUM_GBUFFER_NODES_PER_PIXEL 5

// Maximum number of nodes per allocator tile
#define ALLOCATOR_MAX_NODES_PER_TILE ALLOCATOR_TILE_WIDTH * ALLOCATOR_TILE_HEIGHT * NUM_GBUFFER_NODES_PER_PIXEL

// Constructor
Renderer::Renderer()
    : m_allocatorTileWidth(ALLOCATOR_TILE_WIDTH)
    , m_allocatorTileHeight(ALLOCATOR_TILE_HEIGHT)
    , m_allocatorMaxNodesPerTile(ALLOCATOR_MAX_NODES_PER_TILE)
    , m_allocatorWidth(0)
    , m_allocatorHeight(0)
    , m_lightCount(0)
    , m_lightRotation(0)
    , m_genGBufferProgram(0)
    , m_sortGBufferProgram(0)
    , m_renderGBufferProgram(0)
    , m_width(0)
    , m_height(0)
    , m_enableGen(true)
    , m_enableSort(true)
    , m_enableShade(true)
{
}

// Init
#define CHECKED_INIT(_methodCall) if (!_methodCall) return false;
bool Renderer::Init(GLuint width, GLuint height, GLuint meshesInX, GLuint meshesInY, GLuint meshesInZ)
{
    // Save initialization parameters
    m_width = width;
    m_height = height;
    m_meshesInX = meshesInX;
    m_meshesInY = meshesInY;
    m_meshesInZ = meshesInZ;

    CHECKED_INIT(InitGBuffer());
    CHECKED_INIT(InitMaterials());
    CHECKED_INIT(InitLights());
    CHECKED_INIT(InitCamera());
    CHECKED_INIT(InitGeometry());
    CHECKED_INIT(InitPrograms());
    CHECKED_INIT(InitGL());

    return true;
}

// PreRender
void Renderer::PreRender()
{
    // Transform lights to eye space and populate the buffer
    for (unsigned int light = 0; light < m_lightCount; light++) {
        glm::mat4x4 rotationMatrix;
        rotationMatrix = glm::rotate(rotationMatrix, m_lightRotation, glm::vec3(0.0f, 0.0f, 1.0f));
        glm::vec4 worldPosition = glm::vec4(m_worldSpaceLights[light].eyePosition[0], m_worldSpaceLights[light].eyePosition[1], m_worldSpaceLights[light].eyePosition[2], 1.0f);
        worldPosition = rotationMatrix * worldPosition;            
        glm::vec4 eyePosition = m_viewMatrix * worldPosition;

        // Update eye space lights
        m_eyeSpaceLights[light].eyePosition[0] = eyePosition.x;
        m_eyeSpaceLights[light].eyePosition[1] = eyePosition.y;
        m_eyeSpaceLights[light].eyePosition[2] = eyePosition.z;
    }
    glNamedBufferSubDataEXT(m_lights.Buffer(), 0, m_lightCount * sizeof(Light), &m_eyeSpaceLights[0]);
}

// PostRender
void Renderer::PostRender()
{
    // Update rotation angle of lights by two degrees
    m_lightRotation += 2;
}

// Not handled by glew yet
#define GL_SHADER_GLOBAL_ACCESS_BARRIER_BIT_NV 0x00000010

// Render
void Renderer::Render()
{
    nvtxRangePush(L"Frame");

    // Clear color and depth buffers
    glClearColor(0.0f, 0.0f, 1.0f, 1.0f);
    glClearDepth(1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    nvtxRangePush(L"Gen G-Buffer");
    GenGBuffer();
    glMemoryBarrier(GL_SHADER_GLOBAL_ACCESS_BARRIER_BIT_NV);
    nvtxRangePop();

    nvtxRangePush(L"Sort G-Buffer");
    SortGBuffer();
    glMemoryBarrier(GL_SHADER_GLOBAL_ACCESS_BARRIER_BIT_NV);
    nvtxRangePop();

    nvtxRangePush(L"Shade G-Buffer");
    RenderGBuffer();
    glMemoryBarrier(GL_SHADER_GLOBAL_ACCESS_BARRIER_BIT_NV);
    nvtxRangePop();

    nvtxRangePop();
}

// SetRenderMode
void Renderer::SetRenderMode(RenderMode renderMode)
{
    // Disable all stages to start with
    m_enableGen = false;
    m_enableSort = false;
    m_enableShade = false;

    switch (renderMode) {
    default:
    case RM_DEFAULT:
        m_enableGen = true;
        m_enableSort = true;
        m_enableShade = true;
        break;
    case RM_GEN_ONLY:
        m_enableGen = true;
        break;
    case RM_SORT_ONLY:
        m_enableSort = true;
        break;
    case RM_SHADE_ONLY:
        m_enableShade = true;
        break;
    case RM_GEN_AND_SORT:
        m_enableGen = true;
        m_enableSort = true;
        break;
    case RM_SORT_AND_SHADE:
        m_enableSort = true;
        m_enableShade = true;
        break;
    }
}

// GenGBuffer
void Renderer::GenGBuffer()
{
    // Bail if not generating
    if (!m_enableGen) {
        return;
    }

    // Setup pipeline state
    GLboolean showResult = m_enableSort ? GL_FALSE : GL_TRUE;
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glColorMask(showResult, showResult, showResult, showResult);

    // Use the right program
    glUseProgram(m_genGBufferProgram);

    // Set uniforms for the fragment shader
    glUniform1ui64NV(UNIFORM_GBUFFER_MEMORY_POOL, m_gBufferMemoryPool.GpuAddress());
    glUniform1ui64NV(UNIFORM_GBUFFER_ALLOCATOR, m_gBufferAllocator.GpuAddress());
    glUniform1ui64NV(UNIFORM_GBUFFER, m_gBuffer.GpuAddress());
    glUniformui64NV(UNIFORM_MATERIALS, m_materials.GpuAddress());
    glUniform2ui(UNIFORM_RT_DIMS, m_width, m_height);
    glUniform4ui(UNIFORM_GBUFFER_ALLOCATOR_DIMS, m_allocatorWidth, m_allocatorHeight, m_allocatorTileWidth, m_allocatorTileHeight);
    glUniform1ui64NV(UNIFORM_GBUFFER_ALLOCATOR_MAX_NODES_PER_TILE, m_allocatorMaxNodesPerTile);

    // Reset the allocator
    glNamedCopyBufferSubDataEXT(m_gBufferAllocatorZeroed.Buffer(), m_gBufferAllocator.Buffer(), 0, 0, m_gBufferAllocator.Size());

    // Set all the head pointers in the g buffer to NULL
    glNamedCopyBufferSubDataEXT(m_gBufferZeroed.Buffer(), m_gBuffer.Buffer(), 0, 0, m_gBuffer.Size());

    // Make sure my initialization has finised before I start drawing
    glMemoryBarrier(GL_ALL_BARRIER_BITS);

    // Setup matrices
    glUniform1ui64NV(UNIFORM_MVP_XFORM, m_mvpXforms.GpuAddress());
    glUniform1ui64NV(UNIFORM_MV_XFORM, m_mvXforms.GpuAddress());
    glUniform1ui64NV(UNIFORM_NORMAL_XFORM, m_normalXforms.GpuAddress());
    glUniform1ui64NV(UNIFORM_MATERIAL_IDS, m_materialIds.GpuAddress());

    // Setup
    m_mesh.VertexAttribs();
    m_mesh.MakeResident(GL_READ_ONLY);
    m_mesh.Enable();

    // Draw the mesh
    m_mesh.DrawInstanced(m_meshesInX * m_meshesInY * m_meshesInZ);
}

// SortGBuffer
void Renderer::SortGBuffer()
{
    // Bail if not sorting
    if (!m_enableSort) {
        return;
    }

    // Setup pipeline state
    GLboolean showResult = m_enableShade ? GL_FALSE : GL_TRUE;
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glColorMask(showResult, showResult, showResult, showResult);

    // Use the right program
    glUseProgram(m_sortGBufferProgram);

    // Set uniforms for the fragment shader
    glUniform2ui(UNIFORM_RT_DIMS, m_width, m_height);
    glUniform1ui64NV(UNIFORM_GBUFFER, m_gBuffer.GpuAddress());

    // Setup
    m_screenCoveringMesh.VertexAttribs();
    m_screenCoveringMesh.MakeResident(GL_READ_ONLY);
    m_screenCoveringMesh.Enable();

    // Draw the mesh
    m_screenCoveringMesh.Draw();
}

void Renderer::RenderGBuffer()
{
    // Bail if not shading
    if (!m_enableShade) {
        return;
    }

    // Setup pipeline state
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

    // Use the right program
    glUseProgram(m_renderGBufferProgram);

    // Set uniforms for the fragment shader
    glUniform1ui64NV(UNIFORM_GBUFFER, m_gBuffer.GpuAddress());
    glUniform2ui(UNIFORM_RT_DIMS, m_width, m_height);
    glUniform3f(UNIFORM_CLEAR_COLOR, 0.0f, 0.0f, 0.0f);
    glUniform3f(UNIFORM_AMBIENT_LIGHT_COLOR, 0.2f, 0.2f, 0.2f);
    glUniform1ui(UNIFORM_LIGHT_COUNT, m_lightCount);
    glUniformui64NV(UNIFORM_LIGHTS, m_lights.GpuAddress());
    glUniformui64NV(UNIFORM_MATERIALS, m_materials.GpuAddress());

    // Setup
    m_screenCoveringMesh.VertexAttribs();
    m_screenCoveringMesh.MakeResident(GL_READ_ONLY);
    m_screenCoveringMesh.Enable();

    // Draw the mesh
    m_screenCoveringMesh.Draw();

}

// InitGBuffer
bool Renderer::InitGBuffer()
{
    // Sanity check
    if ((m_width == 0) || (m_height == 0)) {
        return false;
    }

    // Allocate g-buffer memory pool.  This pool can be big, so allocate it first.
    const size_t sizeofGBufferNode = 48; // Includes padding
    const size_t numGBufferNodesPerPixel = NUM_GBUFFER_NODES_PER_PIXEL;
    const size_t numGBufferNodes = m_width * m_height * numGBufferNodesPerPixel;
    if (!m_gBufferMemoryPool.CreateAndMakeResident(numGBufferNodes * sizeofGBufferNode, NULL, GL_STATIC_DRAW, GL_READ_WRITE)) {
        return false;
    }

    // Allocate buffer to store current index into the memory pool and one to clear it
    const size_t sizeofAllocatorTile = sizeof(GLuint64);
    m_allocatorWidth = m_width / m_allocatorTileWidth;
    m_allocatorHeight = m_height / m_allocatorTileHeight;
    const size_t numAllocatorTiles = m_allocatorWidth * m_allocatorHeight;
    GLuint64* pAllocatorZero = new GLuint64[numAllocatorTiles * sizeofAllocatorTile];
    memset(pAllocatorZero, 0, numAllocatorTiles * sizeofAllocatorTile);
    if (!m_gBufferAllocator.CreateAndMakeResident(numAllocatorTiles * sizeofAllocatorTile , pAllocatorZero, GL_STATIC_DRAW, GL_READ_WRITE)) {
        return false;
    }
    if (!m_gBufferAllocatorZeroed.CreateAndMakeResident(numAllocatorTiles * sizeofAllocatorTile, pAllocatorZero, GL_STATIC_DRAW, GL_READ_WRITE)) {
        return false;
    }
    delete[] pAllocatorZero;
    pAllocatorZero = 0;

    // Allocate a buffer to act as the g-buffer... Each entry in the buffer is a pointer
    // to the head of a list of fragments associated with an (x, y).  Also allocate one to clear
    // it.
    GLuint64* pGBufferZero = new GLuint64[m_width * m_height];
    memset(pGBufferZero, 0, m_width * m_height * sizeof(GLuint64));
    if (!m_gBuffer.CreateAndMakeResident(m_width * m_height * sizeof(GLuint64), pGBufferZero, GL_STATIC_DRAW, GL_READ_WRITE)) {
        return false;
    }
    if (!m_gBufferZeroed.CreateAndMakeResident(m_width * m_height * sizeof(GLuint64), pGBufferZero, GL_STATIC_DRAW, GL_READ_WRITE)) {
        return false;
    }
    delete[] pGBufferZero;
    pGBufferZero = 0;

    return true;
}

// PopulateCheckerboardTexture
static void PopulateCheckerboardTexture(BindlessTexture& texture, GLint width, GLint height, GLint blocksPerLength, GLubyte nonWhiteLuminance, GLboolean nonWhiteAlpha)
{
    GLubyte* pCheckerboard = new GLubyte[width * height * 4];
    GLint blockWidth = width / blocksPerLength;
    GLint blockHeight = height / blocksPerLength;

    for (GLint y = 0; y < height; y++) {
        for (GLint x = 0; x < width; x++) {
            GLint blockX = x / blockWidth;
            GLint blockY = y / blockHeight;
            GLubyte luminance = ((blockX + blockY) % 2 == 0) ? 0xFF : nonWhiteLuminance;
            GLubyte alpha = ((blockX + blockY) % 2 == 0) ? 0xFF : nonWhiteAlpha;
            pCheckerboard[y * width * 4 + x * 4 + 0] = luminance;
            pCheckerboard[y * width * 4 + x * 4 + 1] = luminance;
            pCheckerboard[y * width * 4 + x * 4 + 2] = luminance;
            pCheckerboard[y * width * 4 + x * 4 + 3] = alpha;
        }
    }

    // Update the texture image and generate mipmaps for it
    texture.SubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, pCheckerboard);
    texture.GenerateMipmap();

    delete[] pCheckerboard;
}

// InitTextures
static bool InitTextures(GLuint materialCount, vector<BindlessTexture>& diffuseTextures)
{
    // One texture per material
    diffuseTextures.resize(materialCount);

    // Generate the same pattern every time
    srand(0);
    GLint blockLengths[] = {2, 4, 8};

    // Create and populate all the textures
    for (GLuint materialIndex = 0; materialIndex < materialCount; materialIndex++) {
        GLubyte nonWhiteLuminance = (GLubyte)(rand() % 0xFF);
        GLubyte nonWhiteAlpha = 0xFF;
        GLuint blocksPerLengthIndex = rand() % (sizeof(blockLengths)/sizeof(GLint));
        if (!diffuseTextures[materialIndex].Create2DAndMakeResident(GL_TEXTURE_2D, 7, GL_RGBA8, 64, 64)) {
            return false;
        }
        PopulateCheckerboardTexture(diffuseTextures[materialIndex], 64, 64, blockLengths[blocksPerLengthIndex], nonWhiteLuminance, nonWhiteAlpha);
    }

    return true;
}

// InitMaterials
bool Renderer::InitMaterials()
{
    // The number of materials is equal to the number of meshes.  This is a crazy test.
    GLuint materialCount = m_meshesInX * m_meshesInY * m_meshesInZ;
    if (materialCount == 0) {
        return false;
    }

    // Initialize textures first
    if (!InitTextures(materialCount, m_diffuseTextures)) {
        return false;
    }

    // Create a buffer to store the materials
    if (!m_materials.CreateAndMakeResident(materialCount * sizeof(Material), NULL, GL_STATIC_READ, GL_READ_ONLY)) {
        return false;
    }

    // Map the buffer
    Material* pMaterials = (Material*)glMapNamedBufferEXT(m_materials.Buffer(), GL_READ_WRITE);
    if (!pMaterials) {
        return false;
    }

    // Write data into the buffer
    for (GLuint materialIndex = 0; materialIndex < materialCount; materialIndex++) {
        Material& material = pMaterials[materialIndex];

        // Zero initialize everything
        memset(&material, 0, sizeof(material));

        // Populate diffuse color based on the mesh position in space.  Give each m
        GLuint x = materialIndex % m_meshesInX;
        GLuint y = (materialIndex / m_meshesInX) % m_meshesInY;
        GLuint z = (materialIndex / (m_meshesInX * m_meshesInY)) % m_meshesInZ;
        material.diffuseColor[0] = (GLfloat)x / (m_meshesInX - 1);
        material.diffuseColor[1] = (GLfloat)y / (m_meshesInY - 1);
        material.diffuseColor[2] = (GLfloat)z / (m_meshesInZ - 1);
        material.diffuseColor[3] = (x % 2 == 0) ? 1.0f : 0.4f;

        // Populate specular color
        material.specularColor[0] = 1.0f;
        material.specularColor[1] = 1.0f;
        material.specularColor[2] = 1.0f;
        material.specularColor[3] = 100.0f; // Shininess

        // Populate diffuse texture handle
        material.diffuseTexture = m_diffuseTextures[materialIndex].Handle();
    }

    // Unmap the buffer
    glUnmapNamedBufferEXT(m_materials.Buffer());

    pMaterials = (Material*)glMapNamedBufferEXT(m_materials.Buffer(), GL_READ_WRITE);
    glUnmapNamedBufferEXT(m_materials.Buffer());

    return true;
}

#define LIGHT(_x, _y, _z, _dr, _dg, _db, _sr, _sg, _sb) \
{   \
    Light l = {{(_x), (_y), (_z)}, 0.0f, {(_dr), (_dg), (_db)}, 0.0f, {(_sr), (_sg), (_sb)}};   \
    m_worldSpaceLights.push_back(l);                                                            \
}

// InitLights
bool Renderer::InitLights()
{
    // Fixed light configuration for now...
    m_lightRotation = 0;
    LIGHT(-5.0f, -5.0f, -5.0f, 0.4f, 0.4f, 0.4f, 1.0f, 1.0f, 1.0f);
    LIGHT(5.0f, -5.0f, -5.0f, 0.4f, 0.4f, 0.4f, 1.0f, 1.0f, 1.0f);
    LIGHT(-5.0f, 5.0f, 5.0f, 0.4f, 0.4f, 0.4f, 1.0f, 1.0f, 1.0f);
    LIGHT(5.0f, 5.0f, 5.0f, 0.4f, 0.4f, 0.4f, 1.0f, 1.0f, 1.0f);
    LIGHT(0.0f, 0.0f, -5.0f, 0.4f, 0.4f, 0.4f, 0.0f, 0.0f, 0.0f);
    m_lightCount = (GLuint)m_worldSpaceLights.size();

    // Initialize eye space lights from world space lights...
    // The world space lights are kind of a waste.  We could just
    // store world space positions.
    m_eyeSpaceLights = m_worldSpaceLights;

    // Lights aren't populated yet...
    return m_lights.CreateAndMakeResident(m_lightCount * sizeof(Light), NULL, GL_DYNAMIC_COPY, GL_READ_ONLY);
}

static bool PopulateBoxMesh(BindlessMesh& mesh)
{
    // Vertices
    //      1----7
    //      | +Y |
    // 1----3----5----7----1
    // | -Z | +X | +Z | +Z |
    // 0----2----4----6----0
    //      | -Y |
    //      0----6

    const GLfloat maxX = 0.5;
    const GLfloat minX = -0.5;
    const GLfloat maxY = 0.5;
    const GLfloat minY = -0.5;
    const GLfloat maxZ = 0.5;
    const GLfloat minZ = -0.5;

    const GLfloat positions[] = {
        minX, minY, minZ,
        minX, maxY, minZ,
        maxX, minY, minZ,
        maxX, maxY, minZ,
        maxX, minY, maxZ,
        maxX, maxY, maxZ,
        minX, minY, maxZ,
        minX, maxY, maxZ,
    };

    const GLfloat uvs[] = {
        0.0f, 0.0f,
        0.0f, 1.0f,
        1.0f, 0.0f,
        1.0f, 1.0f,
    };

    const GLfloat normals[] = {
        0.0f, 0.0f, -1.0f,
        1.0f, 0.0f, 0.0f,
        0.0f, 0.0f, 1.0f,
        -1.0f, 0.0f, 0.0f,
        0.0f, 1.0f, 0.0f,
        0.0f, -1.0f, 0.0f,
    };

    // Vertex structure
    struct Vertex {
        GLfloat position[3];
        GLfloat uv[2];
        GLfloat normal[3];
    };

#define TRIANGLE_POSITION(i) {positions[i*3 + 0], positions[i*3 + 1], positions[i*3 + 2]}
#define TRIANGLE_UV(i) {uvs[i*2 + 0], uvs[i*2 + 1]}
#define TRIANGLE_NORMAL(i) {normals[i*3 + 0], normals[i*3 + 1], normals[i*3 + 2]}
#define TRIANGLE(p0, p1, p2, uv0, uv1, uv2, n) \
    {TRIANGLE_POSITION(p0), TRIANGLE_UV(uv0), TRIANGLE_NORMAL(n)}, \
    {TRIANGLE_POSITION(p1), TRIANGLE_UV(uv1), TRIANGLE_NORMAL(n)}, \
    {TRIANGLE_POSITION(p2), TRIANGLE_UV(uv2), TRIANGLE_NORMAL(n)}

    const Vertex vertices[] = {
        TRIANGLE(0, 2, 1, 0, 2, 1, 0),
        TRIANGLE(2, 1, 3, 2, 1, 3, 0),
        TRIANGLE(2, 4, 3, 0, 2, 1, 1),
        TRIANGLE(4, 3, 5, 2, 1, 3, 1),
        TRIANGLE(4, 6, 5, 0, 2, 1, 2),
        TRIANGLE(6, 5, 7, 2, 1, 3, 2),
        TRIANGLE(6, 0, 7, 0, 2, 1, 3),
        TRIANGLE(0, 7, 1, 2, 1, 3, 3),
        TRIANGLE(3, 5, 1, 0, 2, 1, 4),
        TRIANGLE(5, 1, 7, 2, 1, 3, 4),
        TRIANGLE(0, 6, 2, 0, 2, 1, 5),
        TRIANGLE(6, 2, 4, 2, 1, 3, 5),
    };

#undef TRIANGLE
#undef TRIANGLE_NORMAL
#undef TRIANGLE_UV
#undef TRIANGLE_POSITION

    // Back it with a buffer
    BindlessBuffer* pBuffer = NULL;
    if (!mesh.CreateBindlessBuffer(pBuffer, sizeof(vertices), vertices, GL_STATIC_DRAW)) {
        return false;
    }

    // Configure the mesh
    mesh.Primitive(GL_TRIANGLES, 36);
    mesh.AddAttribute(VERTEX_ATTRIB_POSITION, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), pBuffer->GpuAddress(), pBuffer->Size());
    mesh.AddAttribute(VERTEX_ATTRIB_UV, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), pBuffer->GpuAddress() + offsetof(Vertex, uv), pBuffer->Size() - offsetof(Vertex, uv));
    mesh.AddAttribute(VERTEX_ATTRIB_NORMAL, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), pBuffer->GpuAddress() + offsetof(Vertex, normal), pBuffer->Size() - offsetof(Vertex, normal));

    return true;
}

// InitGeometry
bool Renderer::InitGeometry()
{
    // Create a single box mesh that we can instance
    if (!PopulateBoxMesh(m_mesh)) {
        return false;
    }

    m_mesh.MakeResident(GL_READ_ONLY);

    // Vectors to store temporarily computed data
    vector<GLfloat> mvXforms;
    vector<GLfloat> mvpXforms;
    vector<GLfloat> normalXforms;
    vector<GLuint> materialIds;

    // Size the vectors appropriately
    mvXforms.resize(m_meshesInX * m_meshesInY * m_meshesInZ * 4 * 4);
    mvpXforms.resize(m_meshesInX * m_meshesInY * m_meshesInZ * 4 * 4);
    normalXforms.resize(m_meshesInX * m_meshesInY * m_meshesInZ * 4 * 3);
    materialIds.resize(m_meshesInX * m_meshesInY * m_meshesInZ);

    // Pointers into the vector
    GLfloat* pCurrentMvXform = &mvXforms[0];
    GLfloat* pCurrentMvpXform = &mvpXforms[0];
    GLfloat* pCurrentNormalXform = &normalXforms[0];
    GLuint* pCurrentMaterialId = &materialIds[0];

    // Compute transforms for all the instances.  We do this once.  Obviously we'd need to recompute
    // these any time the camera moved.
    for (GLuint i = 0; i < m_meshesInX; i++) {
        for (GLuint j = 0; j < m_meshesInY; j++) {
            for (GLuint k = 0; k < m_meshesInZ; k++) {
                // Compute position... A box is 1x1 centered on zero.  Distribute them evenly.
                const int meshSize = 1;
                const float meshPadding = 1.25f;
                const float meshSpaceInX = m_meshesInX * meshSize + (m_meshesInX - 1) * meshPadding;
                const float meshSpaceInY = m_meshesInY * meshSize + (m_meshesInY - 1) * meshPadding;
                const float meshSpaceInZ = m_meshesInZ * meshSize + (m_meshesInZ - 1) * meshPadding;

                // Space these out nicely in a grid
                glm::vec3 position(
                    -meshSpaceInX/2.0f + meshSpaceInX * (i / (float(m_meshesInX - 1))),
                    -meshSpaceInY/2.0f + meshSpaceInY * (j / (float(m_meshesInY - 1))),
                    -meshSpaceInZ * (k / (float(m_meshesInZ - 1))));

                // Compute the M matrix for this instance
                glm::mat4x4 m;
                m = glm::translate(m, position);
                m = glm::scale(m, glm::vec3(1.2f, 1.2f, 0.7f)); // Scale the box a bit

                // Compute the MV, MVP, and normal matrices for this mesh
                glm::mat4x4 mv = m_viewMatrix * m;
                glm::mat4x4 mvp = m_projectionMatrix * mv;
                glm::mat3x3 normal = glm::inverse(glm::transpose(glm::mat3(mv)));

                // Copy the matrices into the temporary storage and advance the pointers
                memcpy(pCurrentMvXform, glm::value_ptr(mv), 4 * 4 * sizeof(GLfloat));
                pCurrentMvXform += 4 * 4;
                memcpy(pCurrentMvpXform, glm::value_ptr(mvp), 4 * 4 * sizeof(GLfloat));
                pCurrentMvpXform += 4 * 4;

                // Normals are tricky... glm3x3 is packed tightly, but glsl mat3x3 isn't.
                // It is really 4x3
                memcpy(pCurrentNormalXform, glm::value_ptr(normal), 3 * sizeof(GLfloat));
                pCurrentNormalXform += 4;
                memcpy(pCurrentNormalXform, glm::value_ptr(normal) + 3, 3 * sizeof(GLfloat));
                pCurrentNormalXform += 4;
                memcpy(pCurrentNormalXform, glm::value_ptr(normal) + 3 + 3, 3 * sizeof(GLfloat));
                pCurrentNormalXform += 4;

                // Compute the material id for this instance
                *pCurrentMaterialId++ = k * (m_meshesInX * m_meshesInY) + j * m_meshesInX + i;
            }
        }
    }

    // Create bindless buffers to store the matrices
    m_mvXforms.CreateAndMakeResident(mvXforms.size() * sizeof(GLfloat), &mvXforms[0], GL_STATIC_READ, GL_READ_ONLY);
    m_mvpXforms.CreateAndMakeResident(mvpXforms.size() * sizeof(GLfloat), &mvpXforms[0], GL_STATIC_READ, GL_READ_ONLY);
    m_normalXforms.CreateAndMakeResident(normalXforms.size() * sizeof(GLfloat), &normalXforms[0], GL_STATIC_READ, GL_READ_ONLY);
    m_materialIds.CreateAndMakeResident(materialIds.size() * sizeof(GLuint), &materialIds[0], GL_STATIC_READ, GL_READ_ONLY);

    if (!PopulateScreenCoveringTriangleMesh(m_screenCoveringMesh)) {
        return false;
    }
    m_screenCoveringMesh.MakeResident(GL_READ_ONLY);

    return true;
}

// InitCamera
bool Renderer::InitCamera()
{
    m_viewMatrix = glm::lookAt(glm::vec3(0.0f, 0.0f, 20.0f), glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    m_projectionMatrix = glm::perspective(45.0f, (float)m_width/(float)m_height, 0.25f, 100.0f);

    return true;
}

// InitPrograms
bool Renderer::InitPrograms()
{
    m_genGBufferProgram = GlslProgram::CreateProgramFromFiles("gen_gbuffer_vs.glsl", "gen_gbuffer_fs.glsl");
    if (m_genGBufferProgram == 0) {
        return false;
    }
    
    m_sortGBufferProgram =  GlslProgram::CreateProgramFromFiles("sort_gbuffer_vs.glsl", "sort_gbuffer_fs.glsl");
    if (m_sortGBufferProgram == 0) {
        return false;
    }

    m_renderGBufferProgram =  GlslProgram::CreateProgramFromFiles("render_gbuffer_vs.glsl", "render_gbuffer_fs.glsl");
    if (m_renderGBufferProgram == 0) {
        return false;
    }

    return true;
}

// InitGL
bool Renderer::InitGL()
{
    glViewport(0, 0, m_width, m_height);
    glScissor(0, 0, m_width, m_height);

    return true;
}