#version 420 core
#extension GL_NV_gpu_shader5 : require
#extension GL_NV_shader_buffer_load : require
#extension GL_ARB_explicit_uniform_location : enable

#define FRAG_COLOR	0

#define UNIFORM_GBUFFER 5
#define UNIFORM_RT_DIMS 8

// GBufferNode structure
struct GBufferNode {
    uint64_t nextAddress; // the next address in the list
    vec2 uv; // uv for texture accesses
    vec3 eyePosition; // eye space position
    float alpha; // alpha value (put it here for better packing)
    vec3 eyeNormal; // eye space normal
    uint material; // material index (put it here for better packing)
};

// The "image" of head pointers... entry (width * y + x) is the head pointer for the list for fragment (x,y)
layout (location = UNIFORM_GBUFFER) uniform GBufferNode** gbuffer;

// The width and height of the screen
layout (location = UNIFORM_RT_DIMS) uniform uvec2 rtDims;

layout(location = FRAG_COLOR, index = 0) out vec4 Color;

void main()
{
    // Ignore helper pixels, which can get into an infinite loop since writes are discarded
    if (gl_SampleMaskIn[0] == 0) {
        return;
    }

    // Get the address where the head of the list is stored
    uint64_t* pHeadAddress = (uint64_t*)gbuffer;
    pHeadAddress += (rtDims.x * uint(gl_FragCoord.y)) + uint(gl_FragCoord.x);

    // An odd little sort... This g buffer supports order independent transparency.
    // Chances are "good" that we'll encounter opaque fragments that will terminate
    // the sort.  The basic idea is to perform a selection sort.  At the end of each
    // iteration we can check the most recently selected item and determine if it
    // is opaque.  If it is, then we are done.  To improve performance later on we can
    // terminate the list.
    GBufferNode* pHead = (GBufferNode*)(*pHeadAddress);
    GBufferNode* pSortedHead = (GBufferNode*)0;
    GBufferNode* pSortedTail = (GBufferNode*)0;
    GBufferNode* pUnsortedHead = pHead;

    // Handle empty list
    if (uint64_t(pHead) == 0) {
        discard;
        return;
    }

    // Loop until sorted...
    bool sorted = false;
    do {
        // Select the next node
        GBufferNode* pPreviousToSelection = (GBufferNode*)0;
        GBufferNode* pSelection = pUnsortedHead;
        GBufferNode* pPrevious = pUnsortedHead;
        GBufferNode* pCurrent = (GBufferNode*)pUnsortedHead->nextAddress;
        while (uint64_t(pCurrent) != 0) {
            // Depth is negative in GL, so closer is actually bigger (closer to zero)
            if (pCurrent->eyePosition.z > pSelection->eyePosition.z) {
                pPreviousToSelection = pPrevious;
                pSelection = pCurrent;
            }
            pPrevious = pCurrent;
            pCurrent = (GBufferNode*)pCurrent->nextAddress;
        }

        // Insert the node into the sorted list
        if (uint64_t(pSortedHead) == 0) {
            pSortedHead = pSelection;
            pSortedTail = pSelection;
        } else {
            pSortedTail->nextAddress = uint64_t(pSelection);
            pSortedTail = pSelection;
        }

        // See if we can terminate early
        if (pSelection->alpha > 0.99f) {
            // Sorted, at least as far as we care to sort.  The selected item is opaque, so
            // none of the other values in the list matter.
            sorted = true;
        } else {
            // Remove the selected node from the unsorted list
            if (uint64_t(pPreviousToSelection) != 0) {
                // Just link the node previous to the selected node to selection's next
                pPreviousToSelection->nextAddress = pSelection->nextAddress;
            } else {
                // The selected node is the head of the unsorted list.
                // Make the next node in the unsorted list selection's next.
                pUnsortedHead = (GBufferNode*)pSelection->nextAddress;

                // The list is sorted when there is no more unsorted head
                if (uint64_t(pUnsortedHead) == 0) {
                    sorted = true;
                }
            }
        }
    } while (!sorted);

    // Null terminate the sorted list
    pSortedTail->nextAddress = 0;

    // Once the list is sorted, write the head pointer back to the g buffer
    *pHeadAddress = uint64_t(pSortedHead);

    // For debugging...
    // Write out something so we can see what is going on... uv is a good choice
    Color.rg = pSortedHead->uv;
    // Set blue component to one to differentiate between gen and sort output
    Color.b = 1.0f;
}