// Renderer.h

class Renderer
{
public:
    // Constructor
    Renderer();

    // Initialize
    bool Init(GLuint width, GLuint height, GLuint meshesInX, GLuint meshesInY, GLuint meshesInZ);

    // PreRender
    void PreRender();

    // Render
    void Render();

    // PostRender
    void PostRender();

    // Render modes
    enum RenderMode {
        RM_DEFAULT,
        RM_GEN_ONLY,
        RM_SORT_ONLY,
        RM_SHADE_ONLY,
        RM_GEN_AND_SORT,
        RM_SORT_AND_SHADE,
    };

    // SetRenderMode
    void SetRenderMode(RenderMode renderMode);

private:
    // Rendering helpers
    void GenGBuffer();
    void SortGBuffer();
    void RenderGBuffer();

    // Initialization helpers
    bool InitGBuffer();
    bool InitMaterials();
    bool InitLights();
    bool InitGeometry();
    bool InitCamera();
    bool InitPrograms();
    bool InitGL();

private:
    // G-Buffer data structures.  There are two data structures.  One is the memory pool that 
    // nodes in the G-Buffer are allocated from.  The other is the G-Buffer itself  The G-Buffer
    // itself has one linked list per (x, y).
    BindlessBuffer m_gBufferMemoryPool;
    BindlessBuffer m_gBuffer;

    // To improve GPU L1 cache behavior, suballocate by tile.  This improves the odds that a list,
    // or at least some of it, will reside in the same cache line.  m_allocatorTileWidth and
    // m_allocatorTileHeight are used to control the size of a tile.  If m_allocatorTileWidth and
    // m_allocatorTileHeight are the size of the render target, then there is a single tile, and
    // the algorithm reduces to the single allocator algorithm.  m_allocatorMaxNodesPerTile is the
    // maximum number of nodes that can be allocated per tile.  When a tile is full the suballocator
    // can either discard new nodes (fragments), or it can search through other tiles to try to

    BindlessBuffer m_gBufferAllocator;
    GLuint m_allocatorTileWidth;
    GLuint m_allocatorTileHeight;
    GLuint m_allocatorMaxNodesPerTile;
    GLuint m_allocatorWidth;
    GLuint m_allocatorHeight;

    // Buffers to aid in clearing the allocator and the g buffer
    BindlessBuffer m_gBufferAllocatorZeroed;
    BindlessBuffer m_gBufferZeroed;

    // Materials
    BindlessBuffer m_materials;

    // Diffuse textures
    vector<BindlessTexture> m_diffuseTextures;

    // Lights
    BindlessBuffer m_lights;
    GLuint m_lightCount;
    GLfloat m_lightRotation;
    vector<Light> m_worldSpaceLights;
    vector<Light> m_eyeSpaceLights;

    // Camera matrices
    glm::mat4x4 m_viewMatrix;
    glm::mat4x4 m_projectionMatrix;

    // Programs
    GLuint m_genGBufferProgram;
    GLuint m_sortGBufferProgram;
    GLuint m_renderGBufferProgram;

    // Number of meshes in each dimension
    GLuint m_meshesInX;
    GLuint m_meshesInY;
    GLuint m_meshesInZ;

    // Meshes
    BindlessMesh m_mesh;

    // Instanced matrices
    BindlessBuffer m_mvpXforms;
    BindlessBuffer m_mvXforms;
    BindlessBuffer m_normalXforms;

    // Instanced materials
    BindlessBuffer m_materialIds;

    // Screen covering mesh for full-screen rendering
    BindlessMesh m_screenCoveringMesh;

    // The render target dimensions
    GLuint m_width;
    GLuint m_height;

    // The render mode
    bool m_enableGen;
    bool m_enableSort;
    bool m_enableShade;
};