#version 420 core
#extension GL_NV_gpu_shader5 : require
#extension GL_NV_shader_buffer_load : require
#extension GL_ARB_explicit_uniform_location : enable
#extension GL_NV_bindless_texture : require

#define FRAG_COLOR	0

#define UNIFORM_GBUFFER_MEMORY_POOL 3
#define UNIFORM_GBUFFER_ALLOCATOR 4
#define UNIFORM_GBUFFER 5
#define UNIFORM_GBUFFER_MAX_NODES 6
#define UNIFORM_MATERIALS 7
#define UNIFORM_RT_DIMS 8
#define UNIFORM_GBUFFER_ALLOCATOR_DIMS 14
#define UNIFORM_GBUFFER_ALLOCATOR_MAX_NODES_PER_TILE 15

#define MAX_ALLOCATION_TRIES 20

// Input from the vertex shader
in block
{
    vec3 eyePosition; // eye space position
    vec3 eyeNormal;  // eye space normal
    vec2 uv; // uv (the same for all textures)
    flat uint material; // material index
} In;

// Material structure
struct Material {
    vec4 diffuseColor; // base diffuse color
    vec4 specularColor; // base specular color, shininess is packed in w
    uint64_t diffuseTexture; // texture to modulate diffuse color
    uint64_t specularTexture; // texture to modulate specular color (not used)
    uint64_t normalTexture; // texture to generate a normal (not used)
    uint diffuseTextureHasAlpha; // 1.0 if the diffuse texture has alpha components 
};

// GBufferNode structure that is written by this shader and consumed by later stages
struct GBufferNode {
    uint64_t nextAddress; // the next address in the list
    vec2 uv; // uv for texture accesses
    vec3 eyePosition; // eye space position
    float alpha; // alpha value (put it here for better packing)
    vec3 eyeNormal; // eye space normal
    uint material; // material index (put it here for better packing)
};

// The base pointer to a block of buffer object memory that we can sub-allocate from
layout (location = UNIFORM_GBUFFER_MEMORY_POOL) uniform GBufferNode* gbufferMemoryPool;

// The the gbuffer allocator.  The allocator is tiled to improve caching.
layout (location = UNIFORM_GBUFFER_ALLOCATOR) uniform uint64_t* pGBufferAllocator;

// The dimensions of the gbuffer allocator.  (x, y) is the size of the allocator in tiles.
// (z, w) is the size of a tile in pixels.
layout (location = UNIFORM_GBUFFER_ALLOCATOR_DIMS) uniform uvec4 gbufferAllocatorDims;

// The maximum nodes that can be allocated from a gbuffer tile
layout (location = UNIFORM_GBUFFER_ALLOCATOR_MAX_NODES_PER_TILE) uniform uint64_t gbufferAllocatorMaxNodesPerTile;

// The "image" of head pointers... entry (width * y + x) is the head pointer for the list for fragment (x,y)
layout (location = UNIFORM_GBUFFER) uniform GBufferNode** gbuffer;

// The array of materials
layout (location = UNIFORM_MATERIALS) uniform Material* pMaterials;

// The width and height of the render target
layout (location = UNIFORM_RT_DIMS) uniform uvec2 rtDims;

// The output color
layout(location = FRAG_COLOR, index = 0) out vec4 Color;

uint64_t AllocGBufferNodeAddressHelper(uint x, uint y)
{
    // Figure out which tile this fragment corresponds to
    uint tileX = x / gbufferAllocatorDims.z;
    uint tileY = y / gbufferAllocatorDims.w;

    // Try to allocate from this tile
    uint64_t* pTile = pGBufferAllocator + (tileY * gbufferAllocatorDims.x) + tileX;
    uint64_t tileOffset = atomicAdd(pTile, 1);

    // Fail if there have already been too many fragments allocated from this tile
    if (tileOffset >= gbufferAllocatorMaxNodesPerTile) {
        return 0;
    }

    // Compute the base tile offset
    uint64_t baseTileOffset = ((tileY * gbufferAllocatorDims.x) + tileX) * gbufferAllocatorMaxNodesPerTile;

    GBufferNode* pNode = gbufferMemoryPool + baseTileOffset + tileOffset;
    return uint64_t(pNode);
}

// Allocates a gbuffer node full of garbage
uint64_t AllocGBufferNodeAddress()
{
    uint x = uint(gl_FragCoord.x);
    uint y = uint(gl_FragCoord.y);

    uint64_t address = AllocGBufferNodeAddressHelper(x, y);

    if (address == 0) {
        for (uint tries = 0; tries < MAX_ALLOCATION_TRIES; tries++) {
            // Try again at a new location.  For an original value x, pick a new value
            // x + tp (mod d), where t is tries, p is relatively prime to d, and d
            // is the dimension of the window in x.  Overflow will tend to land in the
            // same "next" tile.  Use values of 31 and 37.  Any values that will be
            // relatively prime to suspected screen resolutions will work.
            x = (x + 31) % rtDims.x;
            y = (y + 37) % rtDims.y;

            address = AllocGBufferNodeAddressHelper(x, y);
            
            if (address != 0) {
                break;
            }
        }
    }

    return address;
}

// Populates a gbuffer node from fragment shader input
void PopulateGBufferNode(inout GBufferNode* pNode)
{
    // Process the material for this node so that we have an alpha value for the fragment
    Material* pMaterial = pMaterials + In.material;
    float alpha = pMaterial->diffuseColor.a;
    if (pMaterial->diffuseTexture != 0) {
        sampler2D tex = sampler2D(pMaterial->diffuseTexture);
        alpha *= texture(tex, In.uv).a;
    }

    // Write out properties of this node
    pNode->eyePosition = In.eyePosition;
    pNode->alpha = alpha;
    pNode->eyeNormal = In.eyeNormal;
    pNode->material = In.material;
    pNode->uv = In.uv;
    pNode->nextAddress = 0;
}

// Inserts a gbuffer node for the current fragment
void InsertGBufferNode(inout GBufferNode* pNode)
{
    // Setup a pointer to the head of the list
    uint64_t* pHeadAddress = (uint64_t*)gbuffer;
    pHeadAddress += (rtDims.x * uint(gl_FragCoord.y)) + uint(gl_FragCoord.x);

    // Insertion is done using an atomic exchange.  At this point in time
    // pHeadAddress points to something.  It could be NULL (initialization 
    // value), or it could be a node.  The insertion works by pointing pHead
    // at this node (atomically).  The exchange returns the old value, which 
    // can then be glued up to pNode->nextAddress.
    pNode->nextAddress = atomicExchange(pHeadAddress, uint64_t(pNode));
}

void main()
{
    // Try to allocate a new node
    uint64_t nodeAddress = AllocGBufferNodeAddress();
    if (nodeAddress == 0) {
        // Buffer overflow!  Discard once this works... draw red for now.
        discard;
        return;
    }

    // Populate the new node
    GBufferNode* pNode = (GBufferNode*)nodeAddress;
    PopulateGBufferNode(pNode);

    // Insert the new node into the list for this fragment
    InsertGBufferNode(pNode);

    // For debugging...
    // Write out something so we can see what is going on... uv is a good choice
    Color.rg = pNode->uv;
}