#version 420 core
#extension GL_NV_gpu_shader5 : require
#extension GL_NV_shader_buffer_load : require
#extension GL_ARB_explicit_uniform_location : enable
#extension GL_NV_bindless_texture : require

#define FRAG_COLOR	0

#define UNIFORM_GBUFFER 5
#define UNIFORM_MATERIALS 7
#define UNIFORM_RT_DIMS 8
#define UNIFORM_CLEAR_COLOR 9
#define UNIFORM_AMBIENT_LIGHT_COLOR 10
#define UNIFORM_LIGHT_COUNT 11
#define UNIFORM_LIGHTS 12

// GBufferNode structure that is written by this shader and consumed by later stages
struct GBufferNode {
    uint64_t nextAddress; // the next address in the list
    vec2 uv; // uv for texture accesses
    vec3 eyePosition; // eye space position
    float alpha; // alpha value (put it here for better packing)
    vec3 eyeNormal; // eye space normal
    uint material; // material index (put it here for better packing)
};

// Light structure
struct Light {
    vec3 eyePosition;
    vec3 diffuse;
    vec3 specular;
};

// Material structure
struct Material {
    vec4 diffuseColor; // base diffuse color
    vec4 specularColor; // base specular color, shininess is packed in w
    uint64_t diffuseTexture; // texture to modulate diffuse color
    uint64_t specularTexture; // texture to modulate specular color (not used)
    uint64_t normalTexture; // texture to generate a normal (not used)
    uint diffuseTextureHasAlpha; // 1.0 if the diffuse texture has alpha components 
};

// The "image" of head pointers... entry (width * y + x) is the head pointer for the list for fragment (x,y)
layout (location = UNIFORM_GBUFFER) uniform GBufferNode** gbuffer;

// The width and height of the screen
layout (location = UNIFORM_RT_DIMS) uniform uvec2 rtDims;

// The clear color
layout (location = UNIFORM_CLEAR_COLOR) uniform vec3 clearColor;

// Lighting uniforms
layout (location = UNIFORM_AMBIENT_LIGHT_COLOR) uniform vec3 ambientLightColor;
layout (location = UNIFORM_LIGHT_COUNT) uniform unsigned int lightCount;
layout (location = UNIFORM_LIGHTS) uniform Light* pLights;

// The array of materials
layout (location = UNIFORM_MATERIALS) uniform Material* pMaterials;

// The output color
layout(location = FRAG_COLOR, index = 0) out vec4 Color;

//#define NO_LIGHTING 1

// Apply phong lighting
vec3 PhongLighting(in GBufferNode* pNode)
{
    // Get a pointer to the material for this node
    Material* pMaterial = pMaterials + pNode->material;

    // Get a pointer to the list of lights
    Light* pLight = pLights;

    // See if we need two sided lighting (for transparant objects)
    bool twoSidedLighting = bool(pNode->alpha < 0.99);

    // Compute diffuse color
    vec3 diffuseColor = pMaterial->diffuseColor.rgb;
    if (pMaterial->diffuseTexture != 0) {
        sampler2D tex = sampler2D(pMaterial->diffuseTexture);
        diffuseColor *= texture(tex, pNode->uv).rgb;
    }

#ifdef NO_LIGHTING
    return diffuseColor;
#endif // NO_LIGHTING

    // Compute specular color (remember shininess is in the alpha channel)
    vec4 specularColor = pMaterial->specularColor;
    if (pMaterial->specularTexture != 0) {
        sampler2D tex = sampler2D(pMaterial->specularTexture);
        specularColor *= texture(tex, pNode->uv);
    }

    // Compute the ambient contribution
    vec3 ambientContribution = diffuseColor * ambientLightColor;

    // Diffuse and specular contributions start at zero
    vec3 diffuseContribution = vec3(0, 0, 0);
    vec3 specularContribution = vec3(0, 0, 0);

    // Loop over lights
    for (unsigned int light = 0; light < lightCount; light++, pLight++) {
        // Compute light vector
        vec3 lightVector = normalize(pLight->eyePosition - pNode->eyePosition);

        // Compute reflection vector
        vec3 reflectionVector = reflect(-lightVector, pNode->eyeNormal);

        // Compute view vector
        vec3 viewVector = -pNode->eyePosition;
    
        // Diffuse contribution
        float diffuseFactor = dot(lightVector, pNode->eyeNormal);
        diffuseFactor = (twoSidedLighting) ? abs(diffuseFactor) : diffuseFactor;
        if (diffuseFactor > 0.0f) {
            diffuseContribution += diffuseColor.rgb * pLight->diffuse * diffuseFactor;
        }

        // Specular contribution
        float specularFactor = dot(reflectionVector, pNode->eyeNormal);
        specularFactor = (twoSidedLighting) ? abs(specularFactor) : specularFactor;
        if (specularFactor > 0.0f) {
            specularContribution += specularColor.rgb * pLight->specular * pow(specularFactor, specularColor.a);
        }
    }

    return ambientContribution + diffuseContribution + specularContribution;
}

void main()
{
    // Get the address where the head of the list is stored
    uint64_t* pHeadAddress = (uint64_t*)gbuffer;
    pHeadAddress += (rtDims.x * uint(gl_FragCoord.y)) + uint(gl_FragCoord.x);
    GBufferNode* pHead = (GBufferNode*)(*pHeadAddress);
    
    // Handle empty list
    if (uint64_t(pHead) == 0) {
        Color.rgb = clearColor;
        return;
    }

    // Loop over the linked list and shade each node
    GBufferNode* pCurrent = pHead;
    Color = vec4(0.0f, 0.0f, 0.0f, 1.0f);
    while (pCurrent) {
        float alpha = Color.a * pCurrent->alpha;

        // Apply phong lighting
        vec3 color = PhongLighting(pCurrent);

        // Accumulate the color
        Color.rgb += color * alpha;
        Color.a -= alpha;

        pCurrent = (GBufferNode*)pCurrent->nextAddress;
    }

    // Blend in the background color
    Color.rgb += clearColor * Color.a;
}