// BindlessBuffer.cpp

#include "stdafx.h"

// Constructor
BindlessMesh::BindlessMesh()
    : m_mode(GL_TRIANGLES)
    , m_vertices(0)
{

}

// Destructor
BindlessMesh::~BindlessMesh()
{
    // Delete all the buffers owned by this mesh
    vector<BindlessBuffer*>::const_iterator itr = m_bindlessBuffers.begin();
    vector<BindlessBuffer*>::const_iterator end = m_bindlessBuffers.end();
    for (; itr != end; ++itr) {
        delete (*itr);
    }
}

// CreateBindlessBuffer
bool BindlessMesh::CreateBindlessBuffer(BindlessBuffer*& pBindlessBuffer, GLsizeiptr size, const GLvoid* data, GLenum usage)
{
    // Create a new buffer
    pBindlessBuffer = new BindlessBuffer;
    if (!pBindlessBuffer->Create(size, data, usage)) {
        delete pBindlessBuffer;
        pBindlessBuffer = NULL;
        return false;
    }

    // Give ownership to the mesh
    m_bindlessBuffers.push_back(pBindlessBuffer);
    return true;
}

// AddAttribute
bool BindlessMesh::AddAttribute(GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, GLuint64EXT address, GLsizeiptr length)
{
    // Can't respecify an attribute without removing it first
    if (m_attribs.find(index) != m_attribs.end()) {
        return false;
    }

    Attrib attrib = {
        size,
        type,
        normalized,
        stride,
        address,
        length,
        GL_FLOAT
    };

    // Add this attribute
    m_attribs.insert(make_pair(index, attrib));
    return true;
}

// AddAttributeI
bool BindlessMesh::AddAttributeI(GLuint index, GLint size, GLenum type, GLsizei stride, GLuint64EXT address, GLsizeiptr length)
{
    // Can't respecify an attribute without removing it first
    if (m_attribs.find(index) != m_attribs.end()) {
        return false;
    }

    Attrib attrib = {
        size,
        type,
        GL_FALSE,
        stride,
        address,
        length,
        GL_INT
    };

    // Add this attribute
    m_attribs.insert(make_pair(index, attrib));
    return true;
}

// RemoveAttribute
bool BindlessMesh::RemoveAttribute(GLuint index)
{
    map<GLuint, Attrib>::iterator itr = m_attribs.find(index);

    // Can't remove an attribute that doesn't exist
    if (itr == m_attribs.end()) {
        return false;
    }

    // Remove this attribute
    m_attribs.erase(itr);
    return true;
}

// VertexAttribs
void BindlessMesh::VertexAttribs() const
{
    map<GLuint, Attrib>::const_iterator itr = m_attribs.begin();
    map<GLuint, Attrib>::const_iterator end = m_attribs.end();
    for (; itr != end; ++itr) {
        // This isn't ideal... Ideally we'd check to make sure that we aren't setting duplicate state
        if (itr->second.baseType == GL_FLOAT) {
            glVertexAttribFormatNV(itr->first, itr->second.size, itr->second.type, itr->second.normalized, itr->second.stride);
        } else {
            glVertexAttribIFormatNV(itr->first, itr->second.size, itr->second.type, itr->second.stride);
        }
        glBufferAddressRangeNV(GL_VERTEX_ATTRIB_ARRAY_ADDRESS_NV, itr->first, itr->second.address, itr->second.length);
    }
}

// MakeResident
void BindlessMesh::MakeResident(GLenum access) const
{
    vector<BindlessBuffer*>::const_iterator itr = m_bindlessBuffers.begin();
    vector<BindlessBuffer*>::const_iterator end = m_bindlessBuffers.end();
    for (; itr != end; ++itr) {
        (*itr)->MakeResident(access);
    }
}

// MakeNonResident
void BindlessMesh::MakeNonResident() const
{
    vector<BindlessBuffer*>::const_iterator itr = m_bindlessBuffers.begin();
    vector<BindlessBuffer*>::const_iterator end = m_bindlessBuffers.end();
    for (; itr != end; ++itr) {
        (*itr)->MakeNonResident();
    }
}

// Enable
void BindlessMesh::Enable() const
{
    glEnableClientState(GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV);
    map<GLuint, Attrib>::const_iterator itr = m_attribs.begin();
    map<GLuint, Attrib>::const_iterator end = m_attribs.end();
    for (; itr != end; ++itr) {
        glEnableVertexAttribArray(itr->first);
    }
}

// Disable
void BindlessMesh::Disable() const
{
    glDisableClientState(GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV);
    map<GLuint, Attrib>::const_iterator itr = m_attribs.begin();
    map<GLuint, Attrib>::const_iterator end = m_attribs.end();
    for (; itr != end; ++itr) {
        glDisableVertexAttribArray(itr->first);
    }
}

// Primitive
void BindlessMesh::Primitive(GLenum mode, GLuint vertices)
{
    m_mode = mode;
    m_vertices = vertices;
}

// Draw
void BindlessMesh::Draw() const
{
    glDrawArrays(m_mode, 0, m_vertices);
}

// DrawInstanced
void BindlessMesh::DrawInstanced(GLuint instanceCount) const
{
    glDrawArraysInstanced(m_mode, 0, m_vertices, instanceCount);
}

glm::mat4x4& BindlessMesh::Transform()
{
    return m_transform;
}

// IsSupported
bool BindlessMesh::IsSupported()
{
    GLint numExtensions = 0;
    glGetIntegerv(GL_NUM_EXTENSIONS, &numExtensions);
    if (numExtensions == 0) {
        return false;
    }

    for (GLint extension = 0; extension < numExtensions; extension++) {
        if (strcmp((const char*)glGetStringi(GL_EXTENSIONS, extension), "GL_NV_vertex_buffer_unified_memory") == 0) {
            return true;
        }
    }

    return false;
}

// PopulateScreenCoveringTriangleMesh
bool PopulateScreenCoveringTriangleMesh(BindlessMesh& mesh)
{
    struct Vertex {
        GLfloat position[2];
        GLfloat uv[2];
    };

#define VERTEX(x, y, u, v) {{x, y}, {u, v}}
    const Vertex vertices[3] = {
        VERTEX(-1.0f, -1.0f, 0.0f, 0.0f),
        VERTEX(3.0f, -1.0f, 2.0f, 0.0f),
        VERTEX(-1.0f, 3.0f, 0.0f, 2.0f),
    };
#undef VERTEX

    // Back it with a buffer
    BindlessBuffer* pBuffer = NULL;
    if (!mesh.CreateBindlessBuffer(pBuffer, sizeof(vertices), vertices, GL_STATIC_DRAW)) {
        return false;
    }

    // Configure the mesh
    mesh.Primitive(GL_TRIANGLES, 3);
    mesh.AddAttribute(VERTEX_ATTRIB_POSITION, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), pBuffer->GpuAddress(), pBuffer->Size());
    mesh.AddAttribute(VERTEX_ATTRIB_UV, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), pBuffer->GpuAddress() + offsetof(Vertex, uv), pBuffer->Size() - offsetof(Vertex, uv));

    return true;
}