// GlslProgram.cpp

#include "stdafx.h"

using namespace std;

#define INFO_LOG_MAX_SIZE 1024

// CreateProgram
GLuint GlslProgram::CreateProgram(const GLchar* vsSource[], GLuint vsSourceLines, const GLchar* fsSource[], GLuint fsSourceLines)
{
    // Create vertex shader
    GLuint vs = CreateShader(GL_VERTEX_SHADER, vsSource, vsSourceLines);
    if (vs == 0) {
        return 0;
    }

    // Create fragment shader
    GLuint fs = CreateShader(GL_FRAGMENT_SHADER, fsSource, fsSourceLines);
    if (fs == 0) {
        return 0;
    }

    return CreateProgramCommon(vs, fs);
}

// CreateProgramFromFiles
GLuint GlslProgram::CreateProgramFromFiles(const char* vsFilename, const char* fsFilename)
{
    // Create vertex shader
    GLuint vs = CreateShaderFromFile(GL_VERTEX_SHADER, vsFilename);
    if (vs == 0) {
        return 0;
    }

    // Create fragment shader
    GLuint fs = CreateShaderFromFile(GL_FRAGMENT_SHADER, fsFilename);
    if (fs == 0) {
        return 0;
    }

    return CreateProgramCommon(vs, fs);
}

// CreateProgramCommon
GLuint GlslProgram::CreateProgramCommon(GLuint vs, GLuint fs)
{
    GLint linkStatus = GL_FALSE;
    GLchar infoLog[INFO_LOG_MAX_SIZE];

    // Create a program
    GLuint program = glCreateProgram();

    // Attach shaders
    glAttachShader(program, vs);
    if (fs) {
        glAttachShader(program, fs);
    }

    // Link the program
    glLinkProgram(program);

    // Detach shaders
    glDetachShader(program, vs);
    if (fs) {
        glDetachShader(program, fs);
    }

    // Handle failed link
    glGetProgramiv(program, GL_LINK_STATUS, &linkStatus);
    if (!linkStatus) {
        // Get the info log for examination while debugging
        glGetProgramInfoLog(program, INFO_LOG_MAX_SIZE, NULL, infoLog);
        glDeleteProgram(program);
        program = 0;
    }

    // Delete the shaders
    glDeleteShader(vs);
    glDeleteShader(fs);

    return program;
}

// CreateShader
GLuint GlslProgram::CreateShader(GLenum type, const GLchar* source[], GLuint sourceLines)
{
    GLint compileStatus = GL_FALSE;
    GLchar infoLog[INFO_LOG_MAX_SIZE];

    GLuint shader = 0;
    shader = glCreateShader(type);
    glShaderSource(shader, sourceLines, source, NULL);
    glCompileShader(shader);
    glGetShaderiv(shader, GL_COMPILE_STATUS, &compileStatus);
    if (!compileStatus) {
        // Get the info log for examination while debugging
        glGetShaderInfoLog(shader, INFO_LOG_MAX_SIZE, NULL, infoLog);
        glDeleteShader(shader);
        return 0;
    }

    return shader;
}

// CreateShaderFromFile
GLuint GlslProgram::CreateShaderFromFile(GLenum type, const char* filename)
{
    vector<string> shaderStrings;

    // Open the shader
    ifstream file(filename, ios::in);
    if (!file.is_open()) {
        return 0;
    }

    // Read lines and add them to the shader
    string line;
    while (!file.eof()) {
        getline(file, line);
        shaderStrings.push_back(line + "\n");
    }

    file.close();

    // Extract pointers to each line (c-string)
    vector<const char*> shaderStringPtrs;
    shaderStringPtrs.resize(shaderStrings.size());
    for (size_t i = 0 ; i < shaderStrings.size(); i++) {
        shaderStringPtrs[i] = shaderStrings[i].c_str();
    }

    return CreateShader(type, &shaderStringPtrs[0], GLuint(shaderStringPtrs.size()));
}